#include "SearchEngine.h"
#include <sstream>
#include <math.h>
#include <algorithm>

SearchEngine* SearchEngine::m_instance = NULL;

const tRow* SearchEngine::getTermDocuments(const tTerm& term) const
{
	if (m_termByDoc.hasKey(term))
		return (const_cast<SearchEngine*>(this))->m_termByDoc[term];
	return NULL;
}

/**
 * Gather the document vector from our terms by documents map
 * Scoring is done by TF.IDF
 * @param docID   - The document vector to return by its ID
 * @param (out) d - The space vector for the document to gather
 */
void SearchEngine::getDocumentVector(UINT docID, tSpaceVector& d) const
{
	d.clear();
	d.vLength = 0; // Accumulator for the vector length
	double N = numIndexedDocs();
	for ( tTermByDocMap::const_iterator iTerm = m_termByDoc.begin(); iTerm != m_termByDoc.end() ; 
		  iTerm++ ) {
		if (iTerm->second->hasKey(docID)) {
			double t_fi = (*iTerm->second)[docID];
			double d_fi = iTerm->second->size();
			double tf_idf = d[iTerm->first] = (1+log(t_fi))*log(N/d_fi);
			d.vLength += tf_idf * tf_idf;
		}
	}
	d.vLength = sqrt(d.vLength); // No d.vLength = |d|
}

double tSpaceVector::operator*(const tSpaceVector& v) const
{
	double dotProduct = 0;

	for (tSpaceVector::const_iterator iTerm = begin() ; iTerm != end() ; iTerm++){
		if (v.hasKey(iTerm->first)) {
			dotProduct += iTerm->second*v.get(iTerm->first);
		}
	}
	return dotProduct;
}


double tSpaceVector::cosine(const tSpaceVector& v) const
{
	double cosResult = (*this)*v / ( vLength * v.vLength );
	ASSERT(cosResult <= 1.001);
	return cosResult;
}


UINT SearchEngine::index(std::istream& s, const tDocURI& docURI)
{
	UINT newDocID;
	// Test if this docURI already exists
	Vector<tDocURI>::iterator oldDocURI = std::find(m_docID2URI.begin(), m_docID2URI.end(), docURI);
	if (oldDocURI != m_docID2URI.end()) {
		newDocID = oldDocURI - m_docID2URI.begin();
		// Remove old indexed data
		removeIndexedDocument(newDocID);
	} else {
		// Add this new doc at the last column of the document vectors
		newDocID = m_docID2URI.size();
		m_docID2URI.push_back(docURI);
	}

	UINT numTerms = 0;
	// Process the words in the documents and update inverse map accordingly
	m_tokenizer->setInput(s);
	tTerm term;
	while (m_tokenizer->getNextTerm(term)) {
		ASSERT(term.size() > 0);
		numTerms++;

		tRow* row;
		if (m_termByDoc.hasKey(term)) {
			// Word already existed, return already allocated row
			row = m_termByDoc[term];
		} else {
			// New row, create new row vector
			row = m_termByDoc[term] = new tRow();
		}
		if (row->hasKey(newDocID)) 
			(*row)[newDocID]++;
		else
			(*row)[newDocID] = 1;
	}

	//printTermByDocMap(std::cout);
	return numTerms;
}


void SearchEngine::removeIndexedDocument(tDocID docID)
{
	// Go over termByDocs
	for (tTermByDocMap::iterator i = m_termByDoc.begin(); i != m_termByDoc.end(); i++) {
		tTerm term = i->first;
		tRow* row = i->second;
		// Remove this docID from the row if exists
		row->erase(docID);
		// We don't mind keeping empty term rows
	}
}

UINT SearchEngine::query(const std::string& queryStr, Vector<tRerankResult>& queryResults) const
{
	Vector<tTerm> parsedQueryTerms;
	return query(queryStr, queryResults, parsedQueryTerms);
}

UINT SearchEngine::query(const std::string& queryStr, Vector<tRerankResult>& queryResults,
						 Vector<tTerm>& parsedQueryTerms) const
{
	// Tokenize the query into a terms array
	std::stringstream queryStream(queryStr + " ");
	m_tokenizer->setInput(queryStream);
	m_tokenizer->getRemainingTokens(parsedQueryTerms);
	// Run the matcher to match documents with the query
	Set<tMatchResult> matchResults;
	UINT numMatched = m_matcher->match(parsedQueryTerms, matchResults);
	// Rerank the results s.t. the first result has the best reranking score
	m_reranker->rerank(parsedQueryTerms, matchResults, queryResults);
	return numMatched;
}

void SearchEngine::printTermByDocMap(std::ostream& s) const
{
	s << "{ ";
	for (tTermByDocMap::const_iterator i = m_termByDoc.begin(); i != m_termByDoc.end(); i++) {
		s << i->first << " => " << *i->second;
		s << " , ";
	}
	s << "}" << std::endl;
}

// String for marking end of collection file
#define EOF_STRING "-EOF-"

bool SearchEngine::storeCollection(const std::string& filename) const
{
	std::ofstream f(filename.c_str());
	if ((!f.is_open()) || f.fail())
		return false;

	// Write docID2URI vector
	f << "# docID2URI mapping: size, URI1, URI2, ..." << std::endl;
	f << m_docID2URI.size() << std::endl;
	for (int i = 0; i < m_docID2URI.size(); i++)
		f << m_docID2URI[i] << std::endl;
	// Write termByDoc map
	f << "# termByDoc map size" << std::endl;
	f << m_termByDoc.size() << std::endl;
	// Write actual map
	f << "# termByDoc map: <term> <numDocsPerTerm> docID1 termCount1 docID2 termCount2, ... " << std::endl;
	for (tTermByDocMap::const_iterator iMap = m_termByDoc.begin();
		 iMap != m_termByDoc.end(); iMap++) {
		const tRow* row = iMap->second;
		tTerm term = iMap->first;
		f << term << " " << row->size() << " ";
		for (tRow::const_iterator iRow = row->begin(); iRow != row->end(); iRow++) {
			f << iRow->first << " " << iRow->second << " ";
		}
		f << std::endl;
	}
	// Write EOF terminator message (to verity load correctness)
	f << "# EOF line (do not remove)" << std::endl;
	f << EOF_STRING << std::endl;
	f.close();

	if (f.fail())
		return false;

	return true;
}
	
bool SearchEngine::loadCollection(const std::string& filename)
{
	std::ifstream f(filename.c_str());
	if (!f.is_open())
		return false;

	int i;
	// Read docID2URI vector
	UINT numDocs;
	f >> remarkEater >> numDocs;
	m_docID2URI.resize(numDocs);
	for (i = 0; i < m_docID2URI.size(); i++) {
		f >> std::ws;
		m_docID2URI[i] = "";
		// Read upto eol
		char ch;
		while ((f.peek() != '\r') && (f.peek() != '\n')) {
			f.get(ch);
			m_docID2URI[i] += ch;
		}
	}

	// Read termByDoc map size
	m_termByDoc.clear();

	UINT numTerms;
	f >> remarkEater >> numTerms;
	// Read termByDoc map
	f >> remarkEater;
	for (i = 0; i < numTerms; i++) {
		f >> std::ws;
		tTerm term;
		f >> term >> numDocs;
		tRow *row = new tRow();
		for (int j = 0; j < numDocs; j++) {
			tDocID docID;
			UINT numTermsInDoc;
			f >> docID >> numTermsInDoc;
			(*row)[docID] = numTermsInDoc;
		}
		m_termByDoc[term] = row;
	}

	// Read EOF terminator message (to verity load correctness)
	std::string eofTerminator;
	f >> remarkEater >> eofTerminator;
	f.close();

	if ((eofTerminator != EOF_STRING) || f.fail()) {
		// Failed, clear 
		m_docID2URI.clear();
		m_termByDoc.clear();
		return false;
	}

	return true;
}