#ifndef _TOKENIZER_H_
#define _TOKENIZER_H_

#include <Common/SEGlobal.h>

/**
 * Base class for tokenizers.
 *
 * Tokenizers are responsible for parsing input text
 * (either query or document) into terms
 */
class Tokenizer {
protected:
	std::istream* m_stream;
	bool m_shouldDelete;
public:
	Tokenizer();

	virtual ~Tokenizer();

	/**
	 * Set the input for the tokenizer (either an istream
	 * or a filename from which to read)
	 */
	virtual void setInput(std::istream& s);
	virtual void setInput(const char* filename);

	/**
	 * Gets the next term from the input.
	 * @returns true if had next term, false otherwise
	 * Note: The definition of a term is in the derived class
	 */
	virtual bool getNextTerm(tTerm& term) = 0;

	/**
	 * Fills the tokens vector with all remaining tokens
	 */
	void getRemainingTokens(Vector<tTerm>& tokensVector);
protected:
	void setInputStream(std::istream* newStream) {
		if (m_shouldDelete) delete m_stream;
		m_stream = newStream;
	}
	
private:
	Tokenizer(const Tokenizer&);

	Tokenizer& operator=(const Tokenizer&);
};

class ChainedTokenizer : public Tokenizer{
protected:
	Tokenizer* m_tokenizer ;

	ChainedTokenizer(Tokenizer* tokenizer) : m_tokenizer(tokenizer) {}
	~ChainedTokenizer() {
		delete m_tokenizer;
	}
public:

	virtual void setInput(std::istream& s) { m_tokenizer->setInput(s); }
	virtual void setInput(const char* filename) { m_tokenizer->setInput(filename);};

	virtual bool getNextTerm(tTerm& term) = 0;
};

#endif /* _TOKENIZER_H_ */